Object subclass:#HTMLParser
	instanceVariableNames:'inStream outStream textScannedSoFar bindings'
	classVariableNames:'AmpersandEscapes'
	poolDictionaries:''
	category:'System-Documentation'
!

!HTMLParser  class methodsFor:'documentation'!

documentation
"
    instances of this class are used to read HTML documents
    and build a collection of markup elements.
    This markup-collection can be displayed using the HTMLDocumentViewer
    or printed by the HTMLDocumentPrinter.

    [author:]
        Claus Gittinger

    [see also:]
        HTMLMarkup
        HTMLDocumentView HTMLDocumentPrinter
"
!

examples
"
    |p in document|

    p := HTMLParser new.
    in := '../../doc/online/english/TOP.html' asFilename readStream.
    document := p parseText:in.
    in close.
    document inspect


    |v document|

    v := HTMLDocumentView new openAndWait.

    v homeDocument:'../../doc/online/english/TOP.html'.


    |top v document|

    top := StandardSystemView extent:200@500.
    v := HVScrollableView for:HTMLDocumentView miniScrollerH:true in:top.
    v origin:0.0@ 0.0 corner:1.0@1.0.
    top openAndWait.

    v homeDocument:'../../doc/online/english/TOP.html'.



    |v document|

    v := HTMLDocumentView new openAndWait.

    document := (HTMLParser new) 
                        parseText:('../../doc/online/english/programming/viewintro.html' 
                                        asFilename readStream).
    v document:document.

"
! !

!HTMLParser  class methodsFor:'initialization'!

initialize
    AmpersandEscapes isNil ifTrue:[
        AmpersandEscapes := IdentityDictionary new.

        #(
            #nbsp  160          "/ non-breakable space

            #lt    $<
            #gt    $>
            #amp   $&
            #quot  $"

            #copy  169          "/ copyright
            #reg   174          "/ registered

            #cent   162         
            #pound  163
            #yen    165
            #brvbar $|
            #sect   167
            #laquo  171
            #raquo  187
            #plusmn 177
            #micro  181
            #middot 183
            #frac14 188
            #frac12 189
            #frac34 190
            #iquest 191
            #div    247

            #deg   176
            #sup1  185
            #sup2  178
            #sup3  179

            #Oslash 216
            #oslash 248
            #aring  229                         
            #Aring  197

            #ccedil 231
            #Ccedil 199

            #atilde 227
            #Atilde 195
            #ntilde 241
            #Ntilde 209
            #otilde 245
            #Otilde 213

            #auml  228
            #Auml  196
            #uuml  252
            #Uuml  220
            #ouml  246   
            #Ouml  214    
            #euml  235   
            #Euml  203    
            #iuml  239   
            #Iuml  207    
            #yuml  255    

            #acirc  226             
            #Acirc  194
            #icirc  238      
            #Icirc  206
            #ecirc  234
            #Ecirc  202
            #ucirc  251
            #Ucirc  219
            #ocirc  244
            #Ocirc  212

            #agrave 224        
            #Agrave 192
            #egrave 232
            #Egrave 200
            #igrave 236
            #Igrave 204
            #ograve 242
            #Ograve 210
            #ugrave 249
            #Ugrave 217

            #aacute 225
            #Aacute 193
            #eacute 233
            #Eacute 201
            #iacute 237
            #Iacute 205
            #oacute 243
            #Oacute 211
            #uacute 250
            #Uacute 218

            #szlig  223
            #aelig  230
            #AElig  198
        ) pairWiseDo:[:key :val |
            |v|

            v := val.
            val isInteger ifTrue:[
                v := Character value:v
            ].
            AmpersandEscapes at:key put:v
        ].
    ]

    "
     AmpersandEscapes := nil.
     HTMLParser initialize
    "

    "Modified: 19.8.1996 / 15:54:47 / stefan"
    "Modified: 9.9.1996 / 14:52:23 / cg"
! !

!HTMLParser methodsFor:'scanning'!

ampersandEscape
    "parse an ampersand escape; the '&' has already been read."

    |next t ch|

    (next := inStream peek) isLetter ifFalse:[
        next ~~ $# ifTrue:[
            textScannedSoFar nextPut:$&.
            ^ self.
        ]
    ].

    t := ''.

    [inStream atEnd not
     and:[(ch := inStream peek) isSeparator not
     and:[ch ~~ $<
     and:[ch ~~ $;]]]] whileTrue:[
        t := t copyWith:ch.
        inStream next.
    ].

    (inStream atEnd not 
    and:[inStream peek == $;]) ifTrue:[
        inStream next
    ] ifFalse:[
        ('HTML: unterminated &-escape: ', t) infoPrintCR.
    ].

    ch := self ampersandEscape:t.
    ch isNil ifTrue:[
        textScannedSoFar nextPut:$&.
        textScannedSoFar nextPutAll:t
    ] ifFalse:[
        textScannedSoFar nextPutAll:ch asString.
    ]

    "
     (HTMLParser new) parseText:'hello &lt; world > - this is easy'  
     (HTMLParser new) parseText:'hello &lt world > - this is easy'  
     (HTMLParser new) parseText:'hello &foo world > - this is easy'  
     (HTMLParser new) parseText:'hello & foo world > - this is easy'  
    "

    "Modified: 15.7.1996 / 08:50:54 / cg"
!

ampersandEscape:aString
    "return a new string, containing the ampersand escape character"

    |key val|

    (aString at:1) == $# ifTrue:[
        ^ Character value:(Number readFrom:(aString copyFrom:2) onError:$? asciiValue)
    ].

    key := aString asSymbolIfInterned.
    key notNil ifTrue:[
        val := AmpersandEscapes at:key ifAbsent:nil.
    ].

    val isNil ifTrue:[
        bindings notNil ifTrue:[
            val := bindings at:key ifAbsent:nil.
        ]
    ].
    val isNil ifTrue:[
        'HTMLParser: unknown character escape: ' errorPrint. aString errorPrintNL.
    ].
    ^ val

    "
     (HTMLParser new) ampersandEscape:'lt'   
     (HTMLParser new) ampersandEscape:'ouml'  
     (HTMLParser new) ampersandEscape:'#32'  
    "

    "Created: 16.5.1996 / 14:27:45 / cg"
    "Modified: 10.7.1996 / 15:45:25 / cg"
    "Modified: 19.8.1996 / 15:49:02 / stefan"
!

finishTextBlock
    "finish a scanned textBlock; add it to the markup list"

    |m s|

    textScannedSoFar notNil ifTrue:[
        s := textScannedSoFar contents.
        s notEmpty ifTrue:[
            m := HTMLMarkupText new.
            m text:s.
            outStream nextPut:m.
        ].
    ].

    textScannedSoFar := nil

    "Modified: 19.7.1996 / 12:05:00 / cg"
!

parseMarkup
    "return a new string, containing the ampersand escape character"

    |in markText m next|

    in := inStream.

    (next := in peek) isLetter ifTrue:[
    ] ifFalse:[
        next == $/ ifTrue:[
        ] ifFalse:[
            next == $!! ifTrue:[
                in next.
                in peek == $- ifTrue:[
                    (in skipThroughAll:'-->') isNil ifTrue:[
                        'HTML: unclosed comment' errorPrintCR
                    ].
                ] ifFalse:[
                    (in skipThroughAll:'>') isNil ifTrue:[
                        'HTML: unclosed comment' errorPrintCR
                    ].
                ].
"/                in upTo:$>.
                ^ self.
            ] ifFalse:[
                'HTML: ''<'' not followed by letter, ''/'' or ''!!'' - taken as text' infoPrintCR.
                textScannedSoFar nextPut:$<.
                ^ self.
            ]
        ]
    ].

    self finishTextBlock.

    markText := in upTo:$>.

    m := HTMLMarkup for:markText.
    outStream nextPut:m.

    "/ special kludge for <pre> ...
    "/ skip followup CR if there is one

    (markText = 'pre'
    or:[markText = 'PRE']) ifTrue:[
        in peek == Character cr ifTrue:[
            in next
        ]
    ].

    textScannedSoFar := nil

    "
     (HTMLParser new) parseText:'hello <b>world</b> this is easy'  
     (HTMLParser new) parseText:'hello<br>world<p> this is easy'    
     (HTMLParser new) parseText:'hello< br>world<p> this is easy'    
    "

    "Created: 16.5.1996 / 14:27:45 / cg"
    "Modified: 19.7.1996 / 12:08:10 / cg"
    "Modified: 5.8.1996 / 18:28:49 / stefan"
!

parseText:aStringOrStream
    "parse some string, return a list of HTMLMarkups."

    |s|

    s := WriteStream on:(Array new:2000).
    self parseText:aStringOrStream into:s.
    ^ s contents


    "
     (HTMLParser new) parseText:'hello world - this is easy'  
     (HTMLParser new) parseText:'hello &lt; world > - this is easy'  
     (HTMLParser new) parseText:'hello <b>world</b> this is easy'  
     (HTMLParser new) parseText:'hello<br>world<p> this is easy'    
     (HTMLParser new) parseText:'hello<ul><li>world<li>foo</ul><p> this is easy'    
     (HTMLParser new) parseText:'<form><input></form><p> this is easy'    

     (HTMLParser new) 
        parseText:('../../doc/online/english/TOP.html' 
                        asFilename contentsOfEntireFile asString)
     (HTMLParser new) 
        parseText:('../../doc/online/english/programming/viewintro.html' 
                        asFilename contentsOfEntireFile asString)
    "

    "Created: 16.5.1996 / 14:27:45 / cg"
    "Modified: 19.7.1996 / 12:18:25 / cg"
!

parseText:aStringOrStream into:aMarkupStream
    "parse some string, append HTMLMarkups to aMarkupStream."

    |in c|

    in := inStream := aStringOrStream readStream.
    outStream := aMarkupStream.

    textScannedSoFar := '' writeStream.

    [(c := in next) notNil ] whileTrue:[
        c == $< ifTrue:[
            self parseMarkup.
            textScannedSoFar isNil ifTrue:[
                textScannedSoFar := '' writeStream
            ]
        ] ifFalse:[
            c == $& ifTrue:[
                self ampersandEscape.
                textScannedSoFar isNil ifTrue:[
                    textScannedSoFar := '' writeStream
                ]
            ] ifFalse:[
                textScannedSoFar nextPut:c
            ]
        ]
    ].
    self finishTextBlock

    "
     (HTMLParser new) parseText:'hello world - this is easy'  
     (HTMLParser new) parseText:'hello &lt; world > - this is easy'  
     (HTMLParser new) parseText:'hello <b>world</b> this is easy'  
     (HTMLParser new) parseText:'hello<br>world<p> this is easy'    
     (HTMLParser new) parseText:'hello<ul><li>world<li>foo</ul><p> this is easy'    
     (HTMLParser new) parseText:'<form><input></form><p> this is easy'    

     (HTMLParser new) 
        parseText:('../../doc/online/english/TOP.html' 
                        asFilename contentsOfEntireFile asString)
     (HTMLParser new) 
        parseText:('../../doc/online/english/programming/viewintro.html' 
                        asFilename contentsOfEntireFile asString)
    "

    "Created: 19.7.1996 / 12:03:28 / cg"
    "Modified: 19.7.1996 / 12:29:05 / cg"
    "Modified: 5.8.1996 / 18:28:58 / stefan"
!

parseText:aStringOrStream withBindings:metaBindings
    "parse some string, return a list of HTMLMarkups.
     Ambersand variables (i.e. &url) are expanded as given in the
     metabindings dictionary. (this seems to be non-standard HTML,
     but is used in hotjava)"

    bindings := metaBindings.
  ^ self parseText:aStringOrStream.

    "
     (HTMLParser new) parseText:'hello world - this is easy'  
     (HTMLParser new) parseText:'hello &lt; world > - this is easy'  
     (HTMLParser new) parseText:'hello <b>world</b> this is easy'  
     (HTMLParser new) parseText:'hello<br>world<p> this is easy'    
     (HTMLParser new) parseText:'hello<ul><li>world<li>foo</ul><p> this is easy'    
     (HTMLParser new) parseText:'<form><input></form><p> this is easy'    

     (HTMLParser new) 
        parseText:('../../doc/online/english/TOP.html' 
                        asFilename contentsOfEntireFile asString)
     (HTMLParser new) 
        parseText:('../../doc/online/english/programming/viewintro.html' 
                        asFilename contentsOfEntireFile asString)
    "

    "Created: 10.7.1996 / 15:27:07 / cg"
    "Modified: 15.7.1996 / 08:50:02 / cg"
! !

!HTMLParser  class methodsFor:'documentation'!

version
    ^ '$Header: /cvs/stx/stx/libbasic3/Attic/HTMLParser.st,v 1.20 1996-09-09 17:02:29 cg Exp $'
! !
HTMLParser initialize!
